# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0


class CommandNames:
    """
    Command name constants for vector search.
    """

    FT_CREATE = "FT.CREATE"
    FT_DROPINDEX = "FT.DROPINDEX"
    FT_LIST = "FT._LIST"
    FT_SEARCH = "FT.SEARCH"
    FT_INFO = "FT.INFO"
    FT_ALIASADD = "FT.ALIASADD"
    FT_ALIASDEL = "FT.ALIASDEL"
    FT_ALIASUPDATE = "FT.ALIASUPDATE"
    FT_EXPLAIN = "FT.EXPLAIN"
    FT_EXPLAINCLI = "FT.EXPLAINCLI"
    FT_AGGREGATE = "FT.AGGREGATE"
    FT_PROFILE = "FT.PROFILE"
    FT_ALIASLIST = "FT._ALIASLIST"


class FtCreateKeywords:
    """
    Keywords used in the FT.CREATE command.
    """

    SCHEMA = "SCHEMA"
    AS = "AS"
    SORTABLE = "SORTABLE"
    UNF = "UNF"
    NO_INDEX = "NOINDEX"
    ON = "ON"
    PREFIX = "PREFIX"
    SEPARATOR = "SEPARATOR"
    CASESENSITIVE = "CASESENSITIVE"
    DIM = "DIM"
    DISTANCE_METRIC = "DISTANCE_METRIC"
    TYPE = "TYPE"
    INITIAL_CAP = "INITIAL_CAP"
    M = "M"
    EF_CONSTRUCTION = "EF_CONSTRUCTION"
    EF_RUNTIME = "EF_RUNTIME"


class FtSearchKeywords:
    """
    Keywords used in the FT.SEARCH command.
    """

    RETURN = "RETURN"
    TIMEOUT = "TIMEOUT"
    PARAMS = "PARAMS"
    LIMIT = "LIMIT"
    COUNT = "COUNT"
    AS = "AS"


class FtAggregateKeywords:
    """
    Keywords used in the FT.AGGREGATE command.
    """

    LIMIT = "LIMIT"
    FILTER = "FILTER"
    GROUPBY = "GROUPBY"
    REDUCE = "REDUCE"
    AS = "AS"
    SORTBY = "SORTBY"
    MAX = "MAX"
    APPLY = "APPLY"
    LOAD = "LOAD"
    TIMEOUT = "TIMEOUT"
    PARAMS = "PARAMS"


class FtProfileKeywords:
    """
    Keywords used in the FT.PROFILE command.
    """

    QUERY = "QUERY"
    LIMITED = "LIMITED"



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import re
from typing import Mapping, cast

import pytest
from glide.async_commands.core import InfoSection
from glide.config import ProtocolVersion, ReadFrom
from glide.constants import OK
from glide.glide_client import GlideClusterClient
from glide.routes import AllNodes, SlotIdRoute, SlotType

from tests.conftest import create_client
from tests.utils.utils import get_first_result


@pytest.mark.asyncio
# @pytest.mark.usefixtures("multiple_replicas_cluster")
class TestAZAffinity:
    async def _get_num_replicas(self, client: GlideClusterClient) -> int:
        info_replicas = get_first_result(
            await client.info([InfoSection.REPLICATION])
        ).decode()
        match = re.search(r"connected_slaves:(\d+)", info_replicas)
        if match:
            return int(match.group(1))
        else:
            raise ValueError(
                "Could not find the number of replicas in the INFO REPLICATION response"
            )

    @pytest.mark.skip_if_version_below("8.0.0")
    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_routing_with_az_affinity_strategy_to_1_replica(
        self,
        request,
        cluster_mode: bool,
        protocol: ProtocolVersion,
    ):
        """Test that the client with az affinity strategy will only route to the 1 replica with the same az"""
        az = "us-east-1a"
        GET_CALLS = 3
        get_cmdstat = "cmdstat_get" + ":" + f"calls={GET_CALLS}"

        client_for_config_set = await create_client(
            request,
            cluster_mode,
            # addresses=multiple_replicas_cluster.nodes_addr,
            protocol=protocol,
            request_timeout=2000,
        )
        assert type(client_for_config_set) is GlideClusterClient
        # Reset the availability zone for all nodes
        await client_for_config_set.custom_command(
            ["CONFIG", "SET", "availability-zone", ""],
            route=AllNodes(),
        )
        assert await client_for_config_set.config_resetstat() == OK

        # 12182 is the slot of "foo"
        await client_for_config_set.custom_command(
            ["CONFIG", "SET", "availability-zone", az],
            route=SlotIdRoute(SlotType.REPLICA, 12182),
        )

        client_for_testing_az = await create_client(
            request,
            cluster_mode,
            protocol=protocol,
            read_from=ReadFrom.AZ_AFFINITY,
            request_timeout=2000,
            client_az=az,
        )
        assert type(client_for_testing_az) is GlideClusterClient
        for _ in range(GET_CALLS):
            await client_for_testing_az.get("foo")

        info_result = cast(
            Mapping[bytes, bytes],
            await client_for_testing_az.info(
                [InfoSection.SERVER, InfoSection.COMMAND_STATS], AllNodes()
            ),
        )

        # Check that only the replica with az has all the GET calls
        matching_entries_count = sum(
            1
            for value in info_result.values()
            if get_cmdstat in value.decode() and az in value.decode()
        )
        assert matching_entries_count == 1

        # Check that the other replicas have no availability zone set
        changed_az_count = sum(
            1
            for node in info_result.values()
            if ("availability_zone" + ":" + f"{az}") in node.decode()
        )
        assert changed_az_count == 1
        await client_for_testing_az.close()
        await client_for_config_set.close()

    @pytest.mark.skip_if_version_below("8.0.0")
    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_routing_by_slot_to_replica_with_az_affinity_strategy_to_all_replicas(
        self,
        request,
        cluster_mode: bool,
        protocol: ProtocolVersion,
    ):
        """
        Test that the client with AZ affinity strategy routes in a round-robin manner to all replicas within the
        specified AZ
        """

        az = "us-east-1a"
        client_for_config_set = await create_client(
            request,
            cluster_mode,
            # addresses=multiple_replicas_cluster.nodes_addr,
            protocol=protocol,
            request_timeout=2000,
        )
        assert type(client_for_config_set) is GlideClusterClient
        assert await client_for_config_set.config_resetstat() == OK
        await client_for_config_set.custom_command(
            ["CONFIG", "SET", "availability-zone", az], AllNodes()
        )

        client_for_testing_az = await create_client(
            request,
            cluster_mode,
            protocol=protocol,
            read_from=ReadFrom.AZ_AFFINITY,
            request_timeout=2000,
            client_az=az,
        )
        assert type(client_for_testing_az) is GlideClusterClient
        azs = cast(
            Mapping[bytes, Mapping[bytes, bytes]],
            await client_for_testing_az.config_get(["availability-zone"], AllNodes()),
        )

        # Check that all replicas have the availability zone set to the az
        assert all((node[b"availability-zone"].decode() == az) for node in azs.values())

        n_replicas = await self._get_num_replicas(client_for_testing_az)
        GET_CALLS = 4 * n_replicas
        get_cmdstat = "cmdstat_get" + ":" + f"calls={GET_CALLS // n_replicas}"

        for _ in range(GET_CALLS):
            await client_for_testing_az.get("foo")

        info_result = cast(
            Mapping[bytes, bytes],
            await client_for_testing_az.info(
                [InfoSection.COMMAND_STATS, InfoSection.SERVER], AllNodes()
            ),
        )

        # Check that all replicas have the same number of GET calls
        matching_entries_count = sum(
            1
            for value in info_result.values()
            if get_cmdstat in value.decode() and az in value.decode()
        )
        assert matching_entries_count == n_replicas

        await client_for_config_set.close()
        await client_for_testing_az.close()

    @pytest.mark.skip_if_version_below("8.0.0")
    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_az_affinity_non_existing_az(
        self,
        request,
        cluster_mode: bool,
        protocol: ProtocolVersion,
    ):
        GET_CALLS = 4

        client_for_testing_az = await create_client(
            request,
            cluster_mode,
            # addresses=multiple_replicas_cluster.nodes_addr,
            protocol=protocol,
            read_from=ReadFrom.AZ_AFFINITY,
            request_timeout=2000,
            client_az="non-existing-az",
        )
        assert type(client_for_testing_az) is GlideClusterClient
        assert await client_for_testing_az.config_resetstat() == OK

        for _ in range(GET_CALLS):
            await client_for_testing_az.get("foo")

        n_replicas = await self._get_num_replicas(client_for_testing_az)
        # We expect the calls to be distributed evenly among the replicas
        get_cmdstat = "cmdstat_get" + ":" + f"calls={GET_CALLS // n_replicas}"

        info_result = cast(
            Mapping[bytes, bytes],
            await client_for_testing_az.info(
                [InfoSection.COMMAND_STATS, InfoSection.SERVER], AllNodes()
            ),
        )

        matching_entries_count = sum(
            1 for value in info_result.values() if get_cmdstat in value.decode()
        )
        assert matching_entries_count == n_replicas

        await client_for_testing_az.close()

    @pytest.mark.skip_if_version_below("8.0.0")
    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_az_affinity_requires_client_az(
        self, request, cluster_mode: bool, protocol: ProtocolVersion
    ):
        """Test that setting read_from to AZ_AFFINITY without client_az raises an error."""
        with pytest.raises(ValueError):
            await create_client(
                request,
                cluster_mode=cluster_mode,
                protocol=protocol,
                read_from=ReadFrom.AZ_AFFINITY,
                request_timeout=2000,
            )

    @pytest.mark.skip_if_version_below("8.0.0")
    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_az_affinity_replicas_and_primary_routes_to_primary(
        self,
        request,
        cluster_mode: bool,
        protocol: ProtocolVersion,
    ):
        """Test that the client with AZ_AFFINITY_REPLICAS_AND_PRIMARY routes to the primary in the same AZ"""
        az = "us-east-1a"
        other_az = "us-east-1b"
        GET_CALLS = 4

        client_for_config_set = None
        client_for_testing_az = None

        try:
            client_for_config_set = await create_client(
                request,
                cluster_mode,
                protocol=protocol,
                request_timeout=2000,
            )

            assert type(client_for_config_set) is GlideClusterClient

            # Reset stats and set all nodes to other_az
            await client_for_config_set.config_resetstat()
            await client_for_config_set.custom_command(
                ["CONFIG", "SET", "availability-zone", other_az],
                AllNodes(),
            )

            # Set primary for slot 12182 to az
            await client_for_config_set.custom_command(
                ["CONFIG", "SET", "availability-zone", az],
                route=SlotIdRoute(SlotType.PRIMARY, 12182),
            )

            # Verify primary AZ
            primary_az = cast(
                Mapping[bytes, bytes],
                await client_for_config_set.custom_command(
                    ["CONFIG", "GET", "availability-zone"],
                    route=SlotIdRoute(SlotType.PRIMARY, 12182),
                ),
            )
            assert (
                primary_az[b"availability-zone"].decode() == az
            ), f"Primary for slot 12182 is not in the expected AZ {az}"

            # Create test client AFTER configuration
            client_for_testing_az = await create_client(
                request,
                cluster_mode,
                protocol=protocol,
                read_from=ReadFrom.AZ_AFFINITY_REPLICAS_AND_PRIMARY,
                client_az=az,
                request_timeout=2000,
            )

            # Perform GET operations
            for i in range(GET_CALLS):
                await client_for_testing_az.get("foo")

            assert type(client_for_testing_az) is GlideClusterClient
            # Collect info from all nodes
            result = await client_for_testing_az.info(
                [
                    InfoSection.SERVER,
                    InfoSection.REPLICATION,
                    InfoSection.COMMAND_STATS,
                ],
                AllNodes(),
            )
            info_result = cast(dict[bytes, bytes], result)

            matching_entries_count = 0
            total_get_calls = 0
            node_info_list = []

            for node, node_info in info_result.items():
                info_str = node_info.decode()
                is_primary = "role:master" in info_str
                az_match = re.search(r"availability_zone:(\S+)", info_str)
                node_az = az_match.group(1) if az_match else ""
                get_calls_match = re.search(r"cmdstat_get:calls=(\d+)", info_str)
                get_calls = int(get_calls_match.group(1)) if get_calls_match else 0

                total_get_calls += get_calls
                node_info_list.append((node_az, is_primary, get_calls))

                if is_primary and node_az == az and get_calls == GET_CALLS:
                    matching_entries_count += 1
                elif node_az != az and get_calls > 0:
                    pytest.fail(f"GET calls found on node not in AZ {az}")

            assert matching_entries_count == 1, (
                "Exactly one primary in AZ should handle all calls. "
                f"Matching entries: {matching_entries_count}, Total GET calls: {total_get_calls}"
            )
            assert (
                total_get_calls == GET_CALLS
            ), f"Total GET calls mismatch, expected {GET_CALLS}, got {total_get_calls}"

        finally:
            if client_for_testing_az:
                await client_for_testing_az.close()
            if client_for_config_set:
                await client_for_config_set.close()



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from glide.config import (
    AdvancedGlideClientConfiguration,
    AdvancedGlideClusterClientConfiguration,
    BackoffStrategy,
    BaseClientConfiguration,
    GlideClientConfiguration,
    GlideClusterClientConfiguration,
    NodeAddress,
    PeriodicChecksManualInterval,
    PeriodicChecksStatus,
    ReadFrom,
)
from glide.protobuf.connection_request_pb2 import ConnectionRequest
from glide.protobuf.connection_request_pb2 import ReadFrom as ProtobufReadFrom
from glide.protobuf.connection_request_pb2 import TlsMode


def test_default_client_config():
    config = BaseClientConfiguration([])
    assert len(config.addresses) == 0
    assert config.read_from.value == ProtobufReadFrom.Primary
    assert config.use_tls is False
    assert config.client_name is None


def test_convert_to_protobuf():
    config = BaseClientConfiguration(
        [NodeAddress("127.0.0.1")],
        use_tls=True,
        read_from=ReadFrom.PREFER_REPLICA,
        client_name="TEST_CLIENT_NAME",
    )
    request = config._create_a_protobuf_conn_request()
    assert isinstance(request, ConnectionRequest)
    assert request.addresses[0].host == "127.0.0.1"
    assert request.addresses[0].port == 6379
    assert request.tls_mode is TlsMode.SecureTls
    assert request.read_from == ProtobufReadFrom.PreferReplica
    assert request.client_name == "TEST_CLIENT_NAME"


def test_periodic_checks_interval_to_protobuf():
    config = GlideClusterClientConfiguration(
        [NodeAddress("127.0.0.1")],
    )
    request = config._create_a_protobuf_conn_request(cluster_mode=True)
    assert not request.HasField("periodic_checks_disabled")
    assert not request.HasField("periodic_checks_manual_interval")

    config.periodic_checks = PeriodicChecksStatus.DISABLED
    request = config._create_a_protobuf_conn_request(cluster_mode=True)
    assert request.HasField("periodic_checks_disabled")

    config.periodic_checks = PeriodicChecksManualInterval(30)
    request = config._create_a_protobuf_conn_request(cluster_mode=True)
    assert request.periodic_checks_manual_interval.duration_in_sec == 30


def test_convert_config_with_azaffinity_to_protobuf():
    az = "us-east-1a"
    config = BaseClientConfiguration(
        [NodeAddress("127.0.0.1")],
        use_tls=True,
        read_from=ReadFrom.AZ_AFFINITY,
        client_az=az,
    )
    request = config._create_a_protobuf_conn_request()
    assert isinstance(request, ConnectionRequest)
    assert request.tls_mode is TlsMode.SecureTls
    assert request.read_from == ProtobufReadFrom.AZAffinity
    assert request.client_az == az


def test_convert_config_with_azaffinity_replicas_and_primary_to_protobuf():
    az = "us-east-1a"
    config = BaseClientConfiguration(
        [NodeAddress("127.0.0.1")],
        use_tls=True,
        read_from=ReadFrom.AZ_AFFINITY_REPLICAS_AND_PRIMARY,
        client_az=az,
    )
    request = config._create_a_protobuf_conn_request()
    assert isinstance(request, ConnectionRequest)
    assert request.tls_mode is TlsMode.SecureTls
    assert request.read_from == ProtobufReadFrom.AZAffinityReplicasAndPrimary
    assert request.client_az == az


def test_connection_timeout_in_protobuf_request():
    connection_timeout = 5000  # in milliseconds
    config = GlideClientConfiguration(
        [NodeAddress("127.0.0.1")],
        advanced_config=AdvancedGlideClientConfiguration(connection_timeout),
    )
    request = config._create_a_protobuf_conn_request()

    assert isinstance(request, ConnectionRequest)
    assert request.connection_timeout == connection_timeout

    config = GlideClusterClientConfiguration(
        [NodeAddress("127.0.0.1")],
        advanced_config=AdvancedGlideClusterClientConfiguration(connection_timeout),
    )
    request = config._create_a_protobuf_conn_request(cluster_mode=True)

    assert isinstance(request, ConnectionRequest)
    assert request.connection_timeout == connection_timeout


def test_reconnect_strategy_in_protobuf_request():
    reconnect_strategy = BackoffStrategy(7, 69, 3, 18)
    config = GlideClientConfiguration(
        [NodeAddress("127.0.0.1")],
        reconnect_strategy=reconnect_strategy,
    )
    request = config._create_a_protobuf_conn_request()

    assert isinstance(request, ConnectionRequest)
    assert (
        request.connection_retry_strategy.number_of_retries
        == reconnect_strategy.num_of_retries
    )
    assert request.connection_retry_strategy.factor == reconnect_strategy.factor
    assert (
        request.connection_retry_strategy.exponent_base
        == reconnect_strategy.exponent_base
    )
    assert (
        request.connection_retry_strategy.jitter_percent
        == reconnect_strategy.jitter_percent
    )

    config = GlideClusterClientConfiguration(
        [NodeAddress("127.0.0.1")],
        reconnect_strategy=reconnect_strategy,
    )
    request = config._create_a_protobuf_conn_request(cluster_mode=True)

    assert isinstance(request, ConnectionRequest)
    assert (
        request.connection_retry_strategy.number_of_retries
        == reconnect_strategy.num_of_retries
    )
    assert request.connection_retry_strategy.factor == reconnect_strategy.factor
    assert (
        request.connection_retry_strategy.exponent_base
        == reconnect_strategy.exponent_base
    )
    assert (
        request.connection_retry_strategy.jitter_percent
        == reconnect_strategy.jitter_percent
    )



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from enum import Enum


class Limit:
    """
    Represents a limit argument for range queries in various commands.

    The `LIMIT` argument is commonly used to specify a subset of results from the matching elements,
    similar to the `LIMIT` clause in SQL (e.g., `SELECT LIMIT offset, count`).

    This class can be utilized in multiple commands that support limit options,
    such as [ZRANGE](https://valkey.io/commands/zrange), [SORT](https://valkey.io/commands/sort/) and others.

    Args:
        offset (int): The starting position of the range, zero based.
        count (int): The maximum number of elements to include in the range.
            A negative count returns all elements from the offset.

    Examples:
        >>> limit = Limit(0, 10)  # Fetch the first 10 elements
        >>> limit = Limit(5, -1)  # Fetch all elements starting from the 5th element
    """

    def __init__(self, offset: int, count: int):
        self.offset = offset
        self.count = count


class OrderBy(Enum):
    """
    Enumeration representing sorting order options.

    This enum is used for the following commands:

        - `SORT`: General sorting in ascending or descending order.
        - `GEOSEARCH`: Sorting items based on their proximity to a center point.
        - `FT.AGGREGATE`: Used in the SortBy clause of the FT.AGGREGATE command.

    """

    ASC = "ASC"
    """
    ASC: Sort in ascending order.
    """

    DESC = "DESC"
    """
    DESC: Sort in descending order.
    """


class ListDirection(Enum):
    """
    Enumeration representing element popping or adding direction for List commands.
    """

    LEFT = "LEFT"
    """
    LEFT: Represents the option that elements should be popped from or added to the left side of a list.
    """

    RIGHT = "RIGHT"
    """
    RIGHT: Represents the option that elements should be popped from or added to the right side of a list.
    """


class ObjectType(Enum):
    """
    Enumeration representing the data types supported by the database.
    """

    STRING = "String"
    """
    Represents a string data type.
    """

    LIST = "List"
    """
    Represents a list data type.
    """

    SET = "Set"
    """
    Represents a set data type.
    """

    ZSET = "ZSet"
    """
    Represents a sorted set data type.
    """

    HASH = "Hash"
    """
    Represents a hash data type.
    """

    STREAM = "Stream"
    """
    Represents a stream data type.
    """



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from __future__ import annotations

from enum import Enum
from typing import Optional

from .glide import Level as internalLevel
from .glide import py_init, py_log


class Level(Enum):
    ERROR = internalLevel.Error
    WARN = internalLevel.Warn
    INFO = internalLevel.Info
    DEBUG = internalLevel.Debug
    TRACE = internalLevel.Trace
    OFF = internalLevel.Off


class Logger:
    """
    A singleton class that allows logging which is consistent with logs from the internal rust core.
    The logger can be set up in 2 ways -
    1. By calling Logger.init, which configures the logger only if it wasn't previously configured.
    2. By calling Logger.set_logger_config, which replaces the existing configuration, and means that new logs will not be
    saved with the logs that were sent before the call.
    If set_logger_config wasn't called, the first log attempt will initialize a new logger with default configuration decided
    by the Rust core.
    """

    _instance = None
    logger_level: internalLevel

    def __init__(self, level: Optional[Level] = None, file_name: Optional[str] = None):
        level_value = level.value if level else None
        Logger.logger_level = py_init(level_value, file_name)

    @classmethod
    def init(cls, level: Optional[Level] = None, file_name: Optional[str] = None):
        """_summary_
        Initialize a logger if it wasn't initialized before - this method is meant to be used when there is no intention to
        replace an existing logger.
        The logger will filter all logs with a level lower than the given level,
        If given a fileName argument, will write the logs to files postfixed with fileName. If fileName isn't provided,
        the logs will be written to the console.
        Args:
            level (Optional[Level]): Set the logger level to one of [ERROR, WARN, INFO, DEBUG, TRACE, OFF].
            If log level isn't provided, the logger will be configured with default configuration decided by the Rust core.
            file_name (Optional[str]):  If provided the target of the logs will be the file mentioned.
            Otherwise, logs will be printed to the console.
            To turn off logging completely, set the level to Level.OFF.
        """
        if cls._instance is None:
            cls._instance = cls(level, file_name)

    @classmethod
    def log(cls, log_level: Level, log_identifier: str, message: str):
        """Logs the provided message if the provided log level is lower then the logger level.

        Args:
            log_level (Level): The log level of the provided message
            log_identifier (str): The log identifier should give the log a context.
            message (str): The message to log.
        """
        if not cls._instance:
            cls._instance = cls(None)
        if not log_level.value.is_lower(Logger.logger_level):
            return
        py_log(log_level.value, log_identifier, message)

    @classmethod
    def set_logger_config(
        cls, level: Optional[Level] = None, file_name: Optional[str] = None
    ):
        """Creates a new logger instance and configure it with the provided log level and file name.

        Args:
            level (Optional[Level]): Set the logger level to one of [ERROR, WARN, INFO, DEBUG, TRACE, OFF].
            If log level isn't provided, the logger will be configured with default configuration decided by the Rust core.
            file_name (Optional[str]):  If provided the target of the logs will be the file mentioned.
            Otherwise, logs will be printed to the console.
            To turn off logging completely, set the level to OFF.
        """
        Logger._instance = Logger(level, file_name)



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from glide.logger import Level, Logger

from tests.conftest import DEFAULT_TEST_LOG_LEVEL
from tests.utils.utils import compare_maps


class TestLogger:
    def test_init_logger(self):
        # The logger is already configured in the conftest file, so calling init again shouldn't modify the log level
        Logger.init(Level.ERROR)
        assert Logger.logger_level == DEFAULT_TEST_LOG_LEVEL.value

    def test_set_logger_config(self):
        Logger.set_logger_config(Level.INFO)
        assert Logger.logger_level == Level.INFO.value
        # Revert to the tests default log level
        Logger.set_logger_config(DEFAULT_TEST_LOG_LEVEL)
        assert Logger.logger_level == DEFAULT_TEST_LOG_LEVEL.value


class TestCompareMaps:
    def test_empty_maps(self):
        map1 = {}
        map2 = {}
        assert compare_maps(map1, map2) is True

    def test_same_key_value_pairs(self):
        map1 = {"a": 1, "b": 2}
        map2 = {"a": 1, "b": 2}
        assert compare_maps(map1, map2) is True

    def test_different_key_value_pairs(self):
        map1 = {"a": 1, "b": 2}
        map2 = {"a": 1, "b": 3}
        assert compare_maps(map1, map2) is False

    def test_different_key_value_pairs_order(self):
        map1 = {"a": 1, "b": 2}
        map2 = {"b": 2, "a": 1}
        assert compare_maps(map1, map2) is False

    def test_nested_maps_same_values(self):
        map1 = {"a": {"b": 1}}
        map2 = {"a": {"b": 1}}
        assert compare_maps(map1, map2) is True

    def test_nested_maps_different_values(self):
        map1 = {"a": {"b": 1}}
        map2 = {"a": {"b": 2}}
        assert compare_maps(map1, map2) is False

    def test_nested_maps_different_order(self):
        map1 = {"a": {"b": 1, "c": 2}}
        map2 = {"a": {"c": 2, "b": 1}}
        assert compare_maps(map1, map2) is False

    def test_arrays_same_values(self):
        map1 = {"a": [1, 2]}
        map2 = {"a": [1, 2]}
        assert compare_maps(map1, map2) is True

    def test_arrays_different_values(self):
        map1 = {"a": [1, 2]}
        map2 = {"a": [1, 3]}
        assert compare_maps(map1, map2) is False

    def test_null_values(self):
        map1 = {"a": None}
        map2 = {"a": None}
        assert compare_maps(map1, map2) is True

    def test_mixed_types_same_values(self):
        map1 = {
            "a": 1,
            "b": {"c": [2, 3]},
            "d": None,
            "e": "string",
            "f": [1, "2", True],
        }
        map2 = {
            "a": 1,
            "b": {"c": [2, 3]},
            "d": None,
            "e": "string",
            "f": [1, "2", True],
        }
        assert compare_maps(map1, map2) is True

    def test_mixed_types_different_values(self):
        map1 = {
            "a": 1,
            "b": {"c": [2, 3]},
            "d": None,
            "e": "string",
            "f": [1, "2", False],
        }
        map2 = {
            "a": 1,
            "b": {"c": [2, 3]},
            "d": None,
            "f": [1, "2", False],
            "e": "string",
        }
        assert compare_maps(map1, map2) is False



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional


class BitmapIndexType(Enum):
    """
    Enumeration specifying if index arguments are BYTE indexes or BIT indexes. Can be specified in `OffsetOptions`,
    which is an optional argument to the `BITCOUNT` command.

    Since: Valkey version 7.0.0.
    """

    BYTE = "BYTE"
    """
    Specifies that indexes provided to `OffsetOptions` are byte indexes.
    """
    BIT = "BIT"
    """
    Specifies that indexes provided to `OffsetOptions` are bit indexes.
    """


class OffsetOptions:
    """
    Represents offsets specifying a string interval to analyze in the `BITCOUNT` command. The offsets are
    zero-based indexes, with `0` being the first index of the string, `1` being the next index and so on.
    The offsets can also be negative numbers indicating offsets starting at the end of the string, with `-1` being
    the last index of the string, `-2` being the penultimate, and so on.

    Attributes:
        start (int): The starting offset index.
        end (Optional[int]): The ending offset index. Optional since Valkey version 8.0.0 and above for the BITCOUNT
            command. If not provided, it will default to the end of the string.
        index_type (Optional[BitmapIndexType]): The index offset type. This option can only be specified if you are
            using Valkey version 7.0.0 or above. Could be either `BitmapIndexType.BYTE` or `BitmapIndexType.BIT`.
            If no index type is provided, the indexes will be assumed to be byte indexes.
    """

    def __init__(
        self,
        start: int,
        end: Optional[int] = None,
        index_type: Optional[BitmapIndexType] = None,
    ):
        self.start = start
        self.end = end
        self.index_type = index_type

    def to_args(self) -> List[str]:
        args = [str(self.start)]
        if self.end:
            args.append(str(self.end))
        if self.index_type is not None:
            args.append(self.index_type.value)

        return args


class BitwiseOperation(Enum):
    """
    Enumeration defining the bitwise operation to use in the `BITOP` command. Specifies the bitwise operation to
    perform between the passed in keys.
    """

    AND = "AND"
    OR = "OR"
    XOR = "XOR"
    NOT = "NOT"


class BitEncoding(ABC):
    """
    Abstract Base Class used to specify a signed or unsigned argument encoding for the `BITFIELD` or `BITFIELD_RO`
    commands.
    """

    @abstractmethod
    def to_arg(self) -> str:
        """
        Returns the encoding as a string argument to be used in the `BITFIELD` or `BITFIELD_RO`
        commands.
        """
        pass


class SignedEncoding(BitEncoding):
    """
    Represents a signed argument encoding. Must be less than 65 bits long.

    Attributes:
        encoding_length (int): The bit size of the encoding.
    """

    #: Prefix specifying that the encoding is signed.
    SIGNED_ENCODING_PREFIX = "i"

    def __init__(self, encoding_length: int):
        self._encoding = f"{self.SIGNED_ENCODING_PREFIX}{str(encoding_length)}"

    def to_arg(self) -> str:
        return self._encoding


class UnsignedEncoding(BitEncoding):
    """
    Represents an unsigned argument encoding. Must be less than 64 bits long.

    Attributes:
        encoding_length (int): The bit size of the encoding.
    """

    #: Prefix specifying that the encoding is unsigned.
    UNSIGNED_ENCODING_PREFIX = "u"

    def __init__(self, encoding_length: int):
        self._encoding = f"{self.UNSIGNED_ENCODING_PREFIX}{str(encoding_length)}"

    def to_arg(self) -> str:
        return self._encoding


class BitFieldOffset(ABC):
    """Abstract Base Class representing an offset for an array of bits for the `BITFIELD` or `BITFIELD_RO` commands."""

    @abstractmethod
    def to_arg(self) -> str:
        """
        Returns the offset as a string argument to be used in the `BITFIELD` or `BITFIELD_RO`
        commands.
        """
        pass


class BitOffset(BitFieldOffset):
    """
    Represents an offset in an array of bits for the `BITFIELD` or `BITFIELD_RO` commands. Must be greater than or
    equal to 0.

    For example, if we have the binary `01101001` with offset of 1 for an unsigned encoding of size 4, then the value
    is 13 from `0(1101)001`.

    Attributes:
        offset (int): The bit index offset in the array of bits.
    """

    def __init__(self, offset: int):
        self._offset = str(offset)

    def to_arg(self) -> str:
        return self._offset


class BitOffsetMultiplier(BitFieldOffset):
    """
    Represents an offset in an array of bits for the `BITFIELD` or `BITFIELD_RO` commands. The bit offset index is
    calculated as the numerical value of the offset multiplied by the encoding value. Must be greater than or equal
    to 0.

    For example, if we have the binary 01101001 with offset multiplier of 1 for an unsigned encoding of size 4, then
    the value is 9 from `0110(1001)`.

    Attributes:
        offset (int): The offset in the array of bits, which will be multiplied by the encoding value to get the
            final bit index offset.
    """

    #: Prefix specifying that the offset uses an encoding multiplier.
    OFFSET_MULTIPLIER_PREFIX = "#"

    def __init__(self, offset: int):
        self._offset = f"{self.OFFSET_MULTIPLIER_PREFIX}{str(offset)}"

    def to_arg(self) -> str:
        return self._offset


class BitFieldSubCommands(ABC):
    """Abstract Base Class representing subcommands for the `BITFIELD` or `BITFIELD_RO` commands."""

    @abstractmethod
    def to_args(self) -> List[str]:
        """
        Returns the subcommand as a list of string arguments to be used in the `BITFIELD` or `BITFIELD_RO` commands.
        """
        pass


class BitFieldGet(BitFieldSubCommands):
    """
    Represents the "GET" subcommand for getting a value in the binary representation of the string stored in `key`.

    Attributes:
        encoding (BitEncoding): The bit encoding for the subcommand.
        offset (BitFieldOffset): The offset in the array of bits from which to get the value.
    """

    #: "GET" subcommand string for use in the `BITFIELD` or `BITFIELD_RO` commands.
    GET_COMMAND_STRING = "GET"

    def __init__(self, encoding: BitEncoding, offset: BitFieldOffset):
        self._encoding = encoding
        self._offset = offset

    def to_args(self) -> List[str]:
        return [self.GET_COMMAND_STRING, self._encoding.to_arg(), self._offset.to_arg()]


class BitFieldSet(BitFieldSubCommands):
    """
    Represents the "SET" subcommand for setting bits in the binary representation of the string stored in `key`.

    Args:
        encoding (BitEncoding): The bit encoding for the subcommand.
        offset (BitOffset): The offset in the array of bits where the value will be set.
        value (int): The value to set the bits in the binary value to.
    """

    #: "SET" subcommand string for use in the `BITFIELD` command.
    SET_COMMAND_STRING = "SET"

    def __init__(self, encoding: BitEncoding, offset: BitFieldOffset, value: int):
        self._encoding = encoding
        self._offset = offset
        self._value = value

    def to_args(self) -> List[str]:
        return [
            self.SET_COMMAND_STRING,
            self._encoding.to_arg(),
            self._offset.to_arg(),
            str(self._value),
        ]


class BitFieldIncrBy(BitFieldSubCommands):
    """
    Represents the "INCRBY" subcommand for increasing or decreasing bits in the binary representation of the
    string stored in `key`.

    Attributes:
        encoding (BitEncoding): The bit encoding for the subcommand.
        offset (BitOffset): The offset in the array of bits where the value will be incremented.
        increment (int): The value to increment the bits in the binary value by.
    """

    #: "INCRBY" subcommand string for use in the `BITFIELD` command.
    INCRBY_COMMAND_STRING = "INCRBY"

    def __init__(self, encoding: BitEncoding, offset: BitFieldOffset, increment: int):
        self._encoding = encoding
        self._offset = offset
        self._increment = increment

    def to_args(self) -> List[str]:
        return [
            self.INCRBY_COMMAND_STRING,
            self._encoding.to_arg(),
            self._offset.to_arg(),
            str(self._increment),
        ]


class BitOverflowControl(Enum):
    """
    Enumeration specifying bit overflow controls for the `BITFIELD` command.
    """

    WRAP = "WRAP"
    """
    Performs modulo when overflows occur with unsigned encoding. When overflows occur with signed encoding, the value
    restarts at the most negative value. When underflows occur with signed encoding, the value restarts at the most
    positive value.
    """
    SAT = "SAT"
    """
    Underflows remain set to the minimum value, and overflows remain set to the maximum value.
    """
    FAIL = "FAIL"
    """
    Returns `None` when overflows occur.
    """


class BitFieldOverflow(BitFieldSubCommands):
    """
    Represents the "OVERFLOW" subcommand that determines the result of the "SET" or "INCRBY" `BITFIELD` subcommands
    when an underflow or overflow occurs.

    Attributes:
        overflow_control (BitOverflowControl): The desired overflow behavior.
    """

    #: "OVERFLOW" subcommand string for use in the `BITFIELD` command.
    OVERFLOW_COMMAND_STRING = "OVERFLOW"

    def __init__(self, overflow_control: BitOverflowControl):
        self._overflow_control = overflow_control

    def to_args(self) -> List[str]:
        return [self.OVERFLOW_COMMAND_STRING, self._overflow_control.value]


def _create_bitfield_args(subcommands: List[BitFieldSubCommands]) -> List[str]:
    args = []
    for subcommand in subcommands:
        args.extend(subcommand.to_args())

    return args


def _create_bitfield_read_only_args(
    subcommands: List[BitFieldGet],
) -> List[str]:
    args = []
    for subcommand in subcommands:
        args.extend(subcommand.to_args())

    return args



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Union

from glide.constants import TEncodable


class StreamTrimOptions(ABC):
    """
    Abstract base class for stream trim options.

    Attributes:
        exact (bool): If `true`, the stream will be trimmed exactly.
            Otherwise the stream will be trimmed in a near-exact manner, which is more efficient.
        threshold (Union[TEncodable, int]): Threshold for trimming.
        method (str): Method for trimming (e.g., MINID, MAXLEN).
        limit (Optional[int]): Max number of entries to be trimmed. Defaults to None.
            Note: If `exact` is set to `True`, `limit` cannot be specified.
    """

    @abstractmethod
    def __init__(
        self,
        exact: bool,
        threshold: Union[TEncodable, int],
        method: str,
        limit: Optional[int] = None,
    ):
        """
        Initialize stream trim options.
        """
        if exact and limit:
            raise ValueError(
                "If `exact` is set to `True`, `limit` cannot be specified."
            )
        self.exact = exact
        self.threshold = threshold
        self.method = method
        self.limit = limit

    def to_args(self) -> List[str]:
        """
        Convert options to arguments for the command.

        Returns:
            List[str]: List of arguments for the command.
        """
        option_args = [
            self.method,
            "=" if self.exact else "~",
            str(self.threshold),
        ]
        if self.limit is not None:
            option_args.extend(["LIMIT", str(self.limit)])
        return option_args


class TrimByMinId(StreamTrimOptions):
    """
    Stream trim option to trim by minimum ID.

    Attributes:
        exact (bool): If `true`, the stream will be trimmed exactly.
            Otherwise the stream will be trimmed in a near-exact manner, which is more efficient.
        threshold (TEncodable): Threshold for trimming by minimum ID.
        limit (Optional[int]): Max number of entries to be trimmed. Defaults to None.
            Note: If `exact` is set to `True`, `limit` cannot be specified.
    """

    def __init__(self, exact: bool, threshold: TEncodable, limit: Optional[int] = None):
        """
        Initialize trim option by minimum ID.
        """
        super().__init__(exact, threshold, "MINID", limit)


class TrimByMaxLen(StreamTrimOptions):
    """
    Stream trim option to trim by maximum length.

    Attributes:
        exact (bool): If `true`, the stream will be trimmed exactly.
            Otherwise the stream will be trimmed in a near-exact manner, which is more efficient.
        threshold (int): Threshold for trimming by maximum length.
        limit (Optional[int]): Max number of entries to be trimmed. Defaults to None.
            Note: If `exact` is set to `True`, `limit` cannot be specified.
    """

    def __init__(self, exact: bool, threshold: int, limit: Optional[int] = None):
        """
        Initialize trim option by maximum length.
        """
        super().__init__(exact, threshold, "MAXLEN", limit)


class StreamAddOptions:
    """
    Options for adding entries to a stream.

    Attributes:
        id (Optional[TEncodable]): ID for the new entry. If set, the new entry will be added with this ID. If not
            specified, '*' is used.
        make_stream (bool, optional): If set to False, a new stream won't be created if no stream matches the given key.
        trim (Optional[StreamTrimOptions]): If set, the add operation will also trim the older entries in the stream.
            See `StreamTrimOptions`.
    """

    def __init__(
        self,
        id: Optional[TEncodable] = None,
        make_stream: bool = True,
        trim: Optional[StreamTrimOptions] = None,
    ):
        """
        Initialize stream add options.
        """
        self.id = id
        self.make_stream = make_stream
        self.trim = trim

    def to_args(self) -> List[TEncodable]:
        """
        Convert options to arguments for the command.

        Returns:
            List[str]: List of arguments for the command.
        """
        option_args: List[TEncodable] = []
        if not self.make_stream:
            option_args.append("NOMKSTREAM")
        if self.trim:
            option_args.extend(self.trim.to_args())
        option_args.append(self.id if self.id else "*")

        return option_args


class StreamRangeBound(ABC):
    """
    Abstract Base Class used in the `XPENDING`, `XRANGE`, and `XREVRANGE` commands to specify the starting and ending
    range bound for the stream search by stream entry ID.
    """

    @abstractmethod
    def to_arg(self) -> TEncodable:
        """
        Returns the stream range bound as a string argument to be used in the `XRANGE` or `XREVRANGE` commands.
        """
        pass


class MinId(StreamRangeBound):
    """
    Stream ID boundary used to specify the minimum stream entry ID. Can be used in the `XRANGE` or `XREVRANGE` commands
    to get the first stream ID.
    """

    MIN_RANGE_VALKEY_API = "-"

    def to_arg(self) -> str:
        return self.MIN_RANGE_VALKEY_API


class MaxId(StreamRangeBound):
    """
    Stream ID boundary used to specify the maximum stream entry ID. Can be used in the `XRANGE` or `XREVRANGE` commands
    to get the last stream ID.
    """

    MAX_RANGE_VALKEY_API = "+"

    def to_arg(self) -> str:
        return self.MAX_RANGE_VALKEY_API


class IdBound(StreamRangeBound):
    """
    Inclusive (closed) stream ID boundary used to specify a range of IDs to search. Stream ID bounds can be complete
    with a timestamp and sequence number separated by a dash ("-"), for example "1526985054069-0". Stream ID bounds can
    also be incomplete, with just a timestamp.

    Attributes:
        stream_id (str): The stream ID.
    """

    @staticmethod
    def from_timestamp(timestamp: int) -> IdBound:
        """
        Creates an incomplete stream ID boundary without the sequence number for a range search.

        Args:
            timestamp (int): The stream ID timestamp.
        """
        return IdBound(str(timestamp))

    def __init__(self, stream_id: TEncodable):
        """
        Creates a stream ID boundary for a range search.
        """
        self.stream_id = stream_id

    def to_arg(self) -> TEncodable:
        return self.stream_id


class ExclusiveIdBound(StreamRangeBound):
    """
    Exclusive (open) stream ID boundary used to specify a range of IDs to search. Stream ID bounds can be complete with
    a timestamp and sequence number separated by a dash ("-"), for example "1526985054069-0". Stream ID bounds can also
    be incomplete, with just a timestamp.

    Since: Valkey version 6.2.0.

    Attributes:
        stream_id (TEncodable): The stream ID.
    """

    EXCLUSIVE_BOUND_VALKEY_API = "("

    @staticmethod
    def from_timestamp(timestamp: int) -> ExclusiveIdBound:
        """
        Creates an incomplete stream ID boundary without the sequence number for a range search.

        Args:
            timestamp (int): The stream ID timestamp.
        """
        return ExclusiveIdBound(str(timestamp))

    def __init__(self, stream_id: TEncodable):
        """
        Creates a stream ID boundary for a range search.
        """
        if isinstance(stream_id, bytes):
            stream_id = stream_id.decode("utf-8")
        self.stream_id = f"{self.EXCLUSIVE_BOUND_VALKEY_API}{stream_id}"

    def to_arg(self) -> TEncodable:
        return self.stream_id


class StreamReadOptions:
    """
    Options for reading entries from streams. Can be used as an optional argument to `XREAD`.

    Attributes:
        block_ms (Optional[int]): If provided, the request will be blocked for the set amount of milliseconds or
            until the server has the required number of entries. Equivalent to `BLOCK` in the Valkey API.
        count (Optional[int]): The maximum number of elements requested. Equivalent to `COUNT` in the Valkey API.
    """

    READ_COUNT_VALKEY_API = "COUNT"
    READ_BLOCK_VALKEY_API = "BLOCK"

    def __init__(self, block_ms: Optional[int] = None, count: Optional[int] = None):
        self.block_ms = block_ms
        self.count = count

    def to_args(self) -> List[TEncodable]:
        """
        Returns the options as a list of string arguments to be used in the `XREAD` command.

        Returns:
            List[TEncodable]: The options as a list of arguments for the `XREAD` command.
        """
        args: List[TEncodable] = []
        if self.block_ms is not None:
            args.extend([self.READ_BLOCK_VALKEY_API, str(self.block_ms)])

        if self.count is not None:
            args.extend([self.READ_COUNT_VALKEY_API, str(self.count)])

        return args


class StreamGroupOptions:
    """
    Options for creating stream consumer groups. Can be used as an optional argument to `XGROUP CREATE`.

    Attributes:
        make_stream (bool): If set to True and the stream doesn't exist, this creates a new stream with a
            length of 0.
        entries_read: (Optional[int]): A value representing the number of stream entries already read by the
            group. This option can only be specified if you are using Valkey version 7.0.0 or above.
    """

    MAKE_STREAM_VALKEY_API = "MKSTREAM"
    ENTRIES_READ_VALKEY_API = "ENTRIESREAD"

    def __init__(self, make_stream: bool = False, entries_read: Optional[int] = None):
        self.make_stream = make_stream
        self.entries_read = entries_read

    def to_args(self) -> List[TEncodable]:
        """
        Returns the options as a list of string arguments to be used in the `XGROUP CREATE` command.

        Returns:
            List[TEncodable]: The options as a list of arguments for the `XGROUP CREATE` command.
        """
        args: List[TEncodable] = []
        if self.make_stream is True:
            args.append(self.MAKE_STREAM_VALKEY_API)

        if self.entries_read is not None:
            args.extend([self.ENTRIES_READ_VALKEY_API, str(self.entries_read)])

        return args


class StreamReadGroupOptions(StreamReadOptions):
    """
    Options for reading entries from streams using a consumer group. Can be used as an optional argument to
    `XREADGROUP`.

    Attributes:
        no_ack (bool): If set, messages are not added to the Pending Entries List (PEL). This is equivalent to
            acknowledging the message when it is read. Equivalent to `NOACK` in the Valkey API.
        block_ms (Optional[int]): If provided, the request will be blocked for the set amount of milliseconds or
            until the server has the required number of entries. Equivalent to `BLOCK` in the Valkey API.
        count (Optional[int]): The maximum number of elements requested. Equivalent to `COUNT` in the Valkey API.
    """

    READ_NOACK_VALKEY_API = "NOACK"

    def __init__(
        self, no_ack=False, block_ms: Optional[int] = None, count: Optional[int] = None
    ):
        super().__init__(block_ms=block_ms, count=count)
        self.no_ack = no_ack

    def to_args(self) -> List[TEncodable]:
        """
        Returns the options as a list of string arguments to be used in the `XREADGROUP` command.

        Returns:
            List[TEncodable]: The options as a list of arguments for the `XREADGROUP` command.
        """
        args: List[TEncodable] = super().to_args()
        if self.no_ack:
            args.append(self.READ_NOACK_VALKEY_API)

        return args


class StreamPendingOptions:
    """
    Options for `XPENDING` that can be used to filter returned items by minimum idle time and consumer name.

    Attributes:
        min_idle_time_ms (Optional[int]): Filters pending entries by their minimum idle time in milliseconds. This
            option can only be specified if you are using Valkey version 6.2.0 or above.
        consumer_name (Optional[TEncodable]): Filters pending entries by consumer name.
    """

    IDLE_TIME_VALKEY_API = "IDLE"

    def __init__(
        self,
        min_idle_time_ms: Optional[int] = None,
        consumer_name: Optional[TEncodable] = None,
    ):
        self.min_idle_time = min_idle_time_ms
        self.consumer_name = consumer_name


class StreamClaimOptions:
    """
    Options for `XCLAIM`.

    Attributes:
        idle (Optional[int]): Set the idle time (last time it was delivered) of the message in milliseconds. If idle
            is not specified, an idle of `0` is assumed, that is, the time count is reset because the message now has a
            new owner trying to process it.
        idle_unix_time (Optional[int]): This is the same as idle but instead of a relative amount of milliseconds,
            it sets the idle time to a specific Unix time (in milliseconds). This is useful in order to rewrite the AOF
            file generating `XCLAIM` commands.
        retry_count (Optional[int]): Set the retry counter to the specified value. This counter is incremented every
            time a message is delivered again. Normally `XCLAIM` does not alter this counter, which is just served to
            clients when the `XPENDING` command is called: this way clients can detect anomalies, like messages that
            are never processed for some reason after a big number of delivery attempts.
        is_force (Optional[bool]): Creates the pending message entry in the PEL even if certain specified IDs are not
            already in the PEL assigned to a different client. However, the message must exist in the stream, otherwise
            the IDs of non-existing messages are ignored.
    """

    IDLE_VALKEY_API = "IDLE"
    TIME_VALKEY_API = "TIME"
    RETRY_COUNT_VALKEY_API = "RETRYCOUNT"
    FORCE_VALKEY_API = "FORCE"
    JUST_ID_VALKEY_API = "JUSTID"

    def __init__(
        self,
        idle: Optional[int] = None,
        idle_unix_time: Optional[int] = None,
        retry_count: Optional[int] = None,
        is_force: Optional[bool] = False,
    ):
        self.idle = idle
        self.idle_unix_time = idle_unix_time
        self.retry_count = retry_count
        self.is_force = is_force

    def to_args(self) -> List[TEncodable]:
        """
        Converts options for `XCLAIM` into a List.

        Returns:
            List[str]: The options as a list of arguments for the `XCLAIM` command.
        """
        args: List[TEncodable] = []
        if self.idle:
            args.append(self.IDLE_VALKEY_API)
            args.append(str(self.idle))

        if self.idle_unix_time:
            args.append(self.TIME_VALKEY_API)
            args.append(str(self.idle_unix_time))

        if self.retry_count:
            args.append(self.RETRY_COUNT_VALKEY_API)
            args.append(str(self.retry_count))

        if self.is_force:
            args.append(self.FORCE_VALKEY_API)

        return args


def _create_xpending_range_args(
    key: TEncodable,
    group_name: TEncodable,
    start: StreamRangeBound,
    end: StreamRangeBound,
    count: int,
    options: Optional[StreamPendingOptions],
) -> List[TEncodable]:
    args = [key, group_name]
    if options is not None and options.min_idle_time is not None:
        args.extend([options.IDLE_TIME_VALKEY_API, str(options.min_idle_time)])

    args.extend([start.to_arg(), end.to_arg(), str(count)])
    if options is not None and options.consumer_name is not None:
        args.append(options.consumer_name)

    return args



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import struct
from typing import List, Tuple, Type

from google.protobuf import message


class ProtobufCodec:
    """
    Codec for protobuf messages.
    All of the Varint functions were copied from [protobuf](https://github.com/protocolbuffers/protobuf)
    """

    @classmethod
    def _decode_varint_32(cls, buffer, pos):
        decoder_func = cls._varint_decoder((1 << 32) - 1, int)
        return decoder_func(buffer, pos)

    @staticmethod
    def _varint_decoder(mask, result_type):
        """Return an encoder for a basic varint value (does not include tag).

        Decoded values will be bitwise-anded with the given mask before being
        returned, e.g. to limit them to 32 bits.  The returned decoder does not
        take the usual "end" parameter -- the caller is expected to do bounds checking
        after the fact (often the caller can defer such checking until later).  The
        decoder returns a (value, new_pos) pair.
        """

        def decode_varint(buffer, pos):
            result = 0
            shift = 0
            while 1:
                b = buffer[pos]
                result |= (b & 0x7F) << shift
                pos += 1
                if not (b & 0x80):
                    result &= mask
                    result = result_type(result)
                    return (result, pos)
                shift += 7
                if shift >= 64:
                    raise message.DecodeError("Too many bytes when decoding varint.")

        return decode_varint

    @staticmethod
    def _varint_encoder():
        """Return an encoder for a basic varint value (does not include tag)."""

        local_int2byte = struct.Struct(">B").pack

        def encode_varint(write, value):
            bits = value & 0x7F
            value >>= 7
            while value:
                write(local_int2byte(0x80 | bits))
                bits = value & 0x7F
                value >>= 7
            return write(local_int2byte(bits))

        return encode_varint

    @classmethod
    def _varint_bytes(cls, value: int) -> bytes:
        """Encode the given integer as a varint and return the bytes."""

        pieces: List[bytes] = []
        func = cls._varint_encoder()
        func(pieces.append, value)
        return b"".join(pieces)

    @classmethod
    def decode_delimited(
        cls,
        read_bytes: bytearray,
        read_bytes_view: memoryview,
        offset: int,
        message_class: Type[message.Message],
    ) -> Tuple[message.Message, int]:
        try:
            msg_len, new_pos = cls._decode_varint_32(read_bytes_view, offset)
        except IndexError:
            # Didn't read enough bytes to decode the varint
            raise PartialMessageException(
                "Didn't read enough bytes to decode the varint"
            )
        required_read_size = new_pos + msg_len
        if required_read_size > len(read_bytes):
            # Recieved only partial response
            raise PartialMessageException("Recieved only a partial response")
        offset = new_pos
        end = offset + msg_len
        msg_buf = read_bytes_view[offset:end]
        offset += msg_len
        message = message_class()
        message.ParseFromString(msg_buf)
        return (message, offset)

    @classmethod
    def encode_delimited(cls, b_arr: bytearray, message: message.Message) -> None:
        bytes_request = message.SerializeToString()
        varint = cls._varint_bytes(len(bytes_request))
        b_arr.extend(varint)
        b_arr.extend(bytes_request)


class PartialMessageException(Exception):
    pass



#!/usr/bin/env python3
# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import argparse
import os
import subprocess
import sys
from pathlib import Path
from shutil import which
from typing import Any, Dict, List, Optional


def find_project_root() -> Path:
    root = Path(__file__).resolve().parent.parent
    print(f"[INFO] Project root determined at: {root}")
    return root


# Constants
PROTO_REL_PATH = "glide-core/src/protobuf"
PYTHON_CLIENT_PATH = "python/python/glide"
VENV_NAME = ".env"
GLIDE_ROOT = find_project_root()
PYTHON_DIR = GLIDE_ROOT / "python"
VENV_DIR = PYTHON_DIR / VENV_NAME
VENV_BIN_DIR = VENV_DIR / "bin"
PYTHON_EXE = VENV_BIN_DIR / "python"


def check_dependencies() -> None:
    print("[INFO] Checking required dependencies...")
    if not which("rustc"):
        print("❌ Error: Rust is not installed.")
        sys.exit(1)
    print(f"[OK] Rust is installed at: {which('rustc')}")

    if not which("protoc"):
        print("❌ Error: protoc is not installed.")
        sys.exit(1)
    print(f"[OK] protoc is installed at: {which('protoc')}")

    version_output: str = subprocess.check_output(["protoc", "--version"], text=True)
    version = version_output.strip().split()[-1]
    major, minor, *_ = map(int, version.split("."))
    print(f"[INFO] Detected protoc version: {version}")
    if major < 3 or (major == 3 and minor < 2):
        print("❌ Error: protoc version must be >= 3.2.0.")
        sys.exit(1)


def prepare_python_env(no_cache: bool = False):
    if not VENV_DIR.exists():
        print("[INFO] Creating new Python virtual environment...")
        run_command(["python3", "-m", "venv", str(VENV_DIR)], label="venv creation")
    else:
        print("[INFO] Using existing Python virtual environment")

    pip_path = VENV_BIN_DIR / "pip"
    install_cmd = [
        str(pip_path),
        "install",
        "-r",
        str(PYTHON_DIR / "dev_requirements.txt"),
    ]
    if no_cache:
        install_cmd.insert(2, "--no-cache-dir")

    print(f"[INFO] Installing dev requirements using: {' '.join(install_cmd)}")
    run_command(install_cmd, label="pip install requirements")

    install_benchmark_cmd = [
        str(pip_path),
        "install",
        "-r",
        str(GLIDE_ROOT / "benchmarks" / "python" / "requirements.txt"),
    ]
    if no_cache:
        install_benchmark_cmd.insert(2, "--no-cache-dir")

    print(
        f"[INFO] Installing benchmark requirements using: {' '.join(install_benchmark_cmd)}"
    )
    run_command(install_benchmark_cmd, label="pip install benchmark requirements")

    print("[OK] Python environment ready")


def get_venv_env() -> Dict[str, str]:
    env = os.environ.copy()
    env["VIRTUAL_ENV"] = str(VENV_DIR)
    env_path = env.get("PATH", "")
    env["PATH"] = f"{VENV_BIN_DIR}{os.pathsep}{env_path}"
    return env


def generate_protobuf_files() -> None:
    proto_src = GLIDE_ROOT / PROTO_REL_PATH
    proto_dst = GLIDE_ROOT / PYTHON_CLIENT_PATH
    proto_files = list(proto_src.glob("*.proto"))

    if not proto_files:
        print(f"[WARN] No Protobuf files found in {proto_src}")
        return

    print(f"[INFO] Generating Python and .pyi files from Protobuf in: {proto_src}")

    mypy_plugin_path = VENV_BIN_DIR / "protoc-gen-mypy"
    if not mypy_plugin_path.exists():
        print("❌ Error: protoc-gen-mypy not found in venv.")
        print(
            "Hint: Try `pip install --requirement python/dev_requirements.txt` again."
        )
        sys.exit(1)

    run_command(
        [
            "protoc",
            f"--plugin=protoc-gen-mypy={mypy_plugin_path}",
            f"-Iprotobuf={proto_src}",
            f"--python_out={proto_dst}",
            f"--mypy_out={proto_dst}",
            *map(str, proto_files),
        ],
        label="protoc generation",
        env=get_venv_env(),
    )

    print(f"[OK] Protobuf files (.py + .pyi) generated at: {proto_dst}")


def build_async_client(release: bool, no_cache: bool = False) -> None:
    print(
        f"[INFO] Building async client in {'release' if release else 'debug'} mode..."
    )
    env = activate_venv(no_cache)
    generate_protobuf_files()

    cmd = [str(PYTHON_EXE), "-m", "maturin", "develop"]
    if release:
        cmd += ["--release", "--strip"]

    run_command(cmd, cwd=PYTHON_DIR, env=env, label="maturin develop")
    print("[OK] Async client build completed")


def run_command(
    cmd: List[str],
    cwd: Optional[Path] = None,
    env: Optional[dict] = None,
    label: Optional[str] = None,
) -> None:
    label = label or cmd[0]
    print(f"[INFO] Running {label}...")

    try:
        subprocess.run(
            cmd,
            cwd=cwd,
            check=True,
            env=env,
            stderr=subprocess.PIPE,  # Only capture stderr
            text=True,
        )
        print(f"[OK] {label} completed successfully.")
    except subprocess.CalledProcessError as e:
        print(f"❌ Error while running {label}: ")
        print(f"Command: {' '.join(e.cmd)}")
        print(f"Exit code: {e.returncode}")
        if e.stderr:
            print("--- STDERR ---")
            print(e.stderr.strip())
        sys.exit(e.returncode)


def activate_venv(no_cache: bool = False) -> Dict[Any, Any]:
    prepare_python_env(no_cache)
    env = os.environ.copy()
    env["VIRTUAL_ENV"] = str(VENV_BIN_DIR.parent)
    env_path = env["PATH"]
    env["PATH"] = f"{VENV_BIN_DIR}:{env_path}"  # noqa: E231
    return env


def run_linters(check_only: bool = False) -> None:
    print("[INFO] Running Python linters...")
    env = activate_venv()
    generate_protobuf_files()

    isort_args = ["isort", ".", "--profile black"]
    black_args = ["black", "."]

    if check_only:
        isort_args.extend(["--check", "--diff"])
        black_args.extend(["--check", "--diff"])

    run_command(isort_args, cwd=PYTHON_DIR, label="isort", env=env)
    run_command(black_args, cwd=PYTHON_DIR, label="black", env=env)
    run_command(["flake8", "."], cwd=PYTHON_DIR, label="flake8", env=env)
    run_command(["mypy", ".."], cwd=PYTHON_DIR, label="mypy", env=env)

    print("[OK] All linters completed successfully.")


def run_tests(extra_args: Optional[List[str]] = None) -> None:
    print("[INFO] Running test suite...")
    env = get_venv_env()

    cmd = ["pytest", "-v", "--asyncio-mode=auto"]
    if extra_args:
        cmd += extra_args

    run_command(cmd, cwd=PYTHON_DIR, label="pytest", env=env)
    print("[OK] Tests completed successfully")


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Dev utility for building, testing, and linting Glide components.",
        epilog="""
Examples:
    python dev.py build                                   # Build the async client in debug mode
    python dev.py build --client async --mode release     # Build the async client in release mode
    python dev.py protobuf                                # Generate Python protobuf files (.py and .pyi)
    python dev.py lint                                    # Run Python linters
    python dev.py test                                    # Run all tests
        """,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    build_parser = subparsers.add_parser("build", help="Build the Python clients")
    build_parser.add_argument(
        "--client", default="async", choices=["async"], help="Which client to build"
    )
    build_parser.add_argument(
        "--mode", choices=["debug", "release"], default="debug", help="Build mode"
    )
    build_parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Install Python dependencies without cache",
    )

    subparsers.add_parser(
        "protobuf", help="Generate Python protobuf files including .pyi stubs"
    )

    lint_parser = subparsers.add_parser("lint", help="Run all Python linters")
    lint_parser.add_argument(
        "--check",
        action="store_true",
        help="Only check code formatting without modifying files",
    )

    test_parser = subparsers.add_parser("test", help="Run all Python tests")
    test_parser.add_argument(
        "--args",
        nargs=argparse.REMAINDER,
        help="Additional arguments to pass to pytest",
    )

    args = parser.parse_args()
    check_dependencies()

    if args.command == "protobuf":
        print("📦 Generating protobuf Python files...")
        activate_venv()
        generate_protobuf_files()

    elif args.command == "lint":
        print("🔍 Running linters...")
        run_linters(check_only=args.check)

    elif args.command == "test":
        print("🧪 Running tests...")
        run_tests(args.args)

    elif args.command == "build":
        release = args.mode == "release"
        no_cache = args.no_cache
        if args.client in ("async"):
            print(f"🛠 Building async client ({args.mode} mode)...")
            build_async_client(release, no_cache)

    print("[✅ DONE] Task completed successfully.")


if __name__ == "__main__":
    main()
